/*
 * xmltv2vdr.h: A plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 */

#ifndef _XMLTV2VDR_H
#define _XMLTV2VDR_H

#define EPGSOURCES "/var/lib/epgsources" // NEVER (!) CHANGE THIS

#include <vdr/plugin.h>
#include "maps.h"
#include "parse.h"
#include "import.h"

#if __GNUC__ > 3
#define UNUSED(v) UNUSED_ ## v __attribute__((unused))
#else
#define UNUSED(x) x
#endif

static const char *VERSION        = "0.1.0pre";
static const char *DESCRIPTION    = trNOOP("Imports xmltv epg into vdr");

#if VDRVERSNUM < 10726 && !EPGHANDLER
class cEpgHandler : public cListObject
{
public:
    cEpgHandler(void) {}
    virtual ~cEpgHandler() {}
    virtual bool IgnoreChannel(const cChannel *UNUSED(Channel))
    {
        return false;
    }
    virtual bool SetShortText(cEvent *UNUSED(Event), const char *UNUSED(ShortText))
    {
        return false;
    }
    virtual bool SetDescription(cEvent *UNUSED(Event), const char *UNUSED(Description))
    {
        return false;
    }
    virtual bool SetContents(cEvent *UNUSED(Event), uchar *UNUSED(Contents))
    {
        return false;
    }
    virtual bool SetParentalRating(cEvent *UNUSED(Event), int UNUSED(ParentalRating))
    {
        return false;
    }
};
#endif

class cEPGSources;

class cImport;

class cEPGHandler : public cEpgHandler
{
private:
    const char *epgfile;
    cEPGMappings *maps;
    cEPGSources *sources;
    cImport *import;
    bool ProcessChannel(const tChannelID ChannelID);
    cEPGSource *GetSource(const char *Name);    
public:
    cEPGHandler(const char *EpgFile, cEPGSources *Sources, cEPGMappings *Maps, cTEXTMappings *Texts);
    virtual ~cEPGHandler();
    virtual bool IgnoreChannel(const cChannel *Channel);
    virtual bool SetShortText(cEvent *Event, const char *ShortText);
    virtual bool SetDescription(cEvent *Event, const char *Description);
    virtual bool SetContents(cEvent *Event, uchar *Contents);
    virtual bool SetParentalRating(cEvent *Event, int ParentalRating);
};

class cEPGChannel : public cListObject
{
private:
    bool inuse;
    const char *name;
public:
    cEPGChannel(const char *Name, bool InUse=false);
    ~cEPGChannel();
    virtual int Compare(const cListObject &ListObject) const;
    bool InUse()
    {
        return inuse;
    }
    void SetUsage(bool InUse)
    {
        inuse=InUse;
    }
    const char *Name()
    {
        return name;
    }
};

class cEPGChannels : public cList<cEPGChannel> {};

class cEPGExecutor;

class cEPGSource : public cListObject
{
private:
    const char *name;
    const char *confdir;
    const char *epgfile;
    const char *pin;
    int loglen;
    cParse *parse;
    cImport *import;
    bool ready2parse;
    bool usepipe;
    bool needpin;
    bool running;
    bool upstartdone;
    int daysinadvance;
    int exec_upstart;
    int exec_weekday;
    int exec_time;
    int daysmax;
    time_t lastexec;
    void add2Log(const char prefix, const char *line);
    bool ReadConfig();
    int ReadOutput(char *&result, size_t &l);
    cEPGChannels channels;
public:
    cEPGSource(const char *Name,const char *ConfDir,const char *EPGFile,
               cEPGMappings *Maps, cTEXTMappings *Texts);
    ~cEPGSource();
    int Execute(cEPGExecutor &myExecutor);
    int Import(cEPGExecutor &myExecutor);
    bool RunItNow();
    time_t NextRunTime();
    void Store(void);
    void ChangeChannelSelection(int *Selection);
    char *Log;
    cEPGChannels *ChannelList()
    {
        return &channels;
    }
    int ExecTime()
    {
        return exec_time;
    }
    int ExecWeekDay()
    {
        return exec_weekday;
    }
    int ExecUpStart()
    {
        return exec_upstart;
    }
    int DaysMax()
    {
        return daysmax;
    }
    int DaysInAdvance()
    {
        return daysinadvance;
    }
    bool NeedPin()
    {
        return needpin;
    }
    const char *EPGFile()
    {
        return epgfile;
    }
    const char *Name()
    {
        return name;
    }
    const char *Pin()
    {
        return pin;
    }
    void ChangeExec(int UpStart, int Time, int WeekDay)
    {
        exec_upstart=UpStart;
        exec_time=Time;
        exec_weekday=WeekDay;
    }
    void ChangeDaysInAdvance(int NewDaysInAdvance)
    {
        daysinadvance=NewDaysInAdvance;
    }
    void ChangePin(const char *NewPin)
    {
        if (pin) free((void *) pin);
        pin=strdup(NewPin);
    }
    time_t LastExecution()
    {
        return lastexec;
    }
    void Dlog(const char *format, ...);
    void Elog(const char *format, ...);
    void Ilog(const char *format, ...);
    bool Active()
    {
        return running;
    }
};

class cEPGSources : public cList<cEPGSource> {};

class cEPGExecutor : public cThread
{
private:
    cEPGSources *sources;
    cTEXTMappings *textmappings;
    bool epall;
public:
    cEPGExecutor(cEPGSources *Sources);
    bool StillRunning()
    {
        return Running();
    }
    void Stop()
    {
        Cancel(3);
    }
    void SetOptions(bool EPAll, cTEXTMappings *TextMappings)
    {
        epall=EPAll;
        textmappings=TextMappings;
    }
    virtual void Action();
};

class cPluginXmltv2vdr : public cPlugin
{
private:
    cEPGHandler *epghandler;
    cEPGExecutor epgexecutor;
    cEPGMappings epgmappings;
    cEPGSources epgsources;
    cTEXTMappings textmappings;
    void removeepgsources();
    void removeepgmappings();
    void removetextmappings();
    time_t nexttime();
    bool runitnow();
    bool epgsourceexists(const char *name);
    time_t last_housetime_t;
    char *confdir;
    char *epgfile;
    bool epall;
    bool wakeup;
public:
    void SetEPAll(bool Value)
    {
        epall=Value;
        epgexecutor.SetOptions(epall,&textmappings);
    }
    bool EPAll()
    {
        return epall;
    }
    void SetWakeUp(bool Value)
    {
      wakeup=Value;
    }
    bool WakeUp()
    {
      return wakeup;
    }
    void ReadInEPGSources(bool Reload=false);
    int EPGSourceCount()
    {
        return epgsources.Count();
    }
    cEPGSource *EPGSource(int Index)
    {
        return epgsources.Get(Index);
    }
    int EPGMappingCount()
    {
        return epgmappings.Count();
    }
    cEPGMapping *EPGMapping(int Index)
    {
        return epgmappings.Get(Index);
    }
    cEPGMapping *EPGMapping(const char *ChannelName);
    void EPGMappingAdd(cEPGMapping *Map)
    {
        epgmappings.Add(Map);
    }
    cTEXTMapping *TEXTMapping(const char *Name);
    void TEXTMappingAdd(cTEXTMapping *TextMap)
    {
        textmappings.Add(TextMap);
    }
    cPluginXmltv2vdr(void);
    virtual ~cPluginXmltv2vdr();
    virtual const char *Version(void)
    {
        return VERSION;
    }
    virtual const char *Description(void)
    {
        return tr(DESCRIPTION);
    }
    virtual const char *CommandLineHelp(void);
    virtual bool ProcessArgs(int argc, char *argv[]);
    virtual bool Initialize(void);
    virtual bool Start(void);
    virtual void Stop(void);
    virtual void Housekeeping(void);
    virtual void MainThreadHook(void);
    virtual cString Active(void);
    virtual time_t WakeupTime(void);
    virtual const char *MainMenuEntry(void);
    virtual cOsdObject *MainMenuAction(void);
    virtual cMenuSetupPage *SetupMenu(void);
    virtual bool SetupParse(const char *Name, const char *Value);
    virtual bool Service(const char *Id, void *Data = NULL);
    virtual const char **SVDRPHelpPages(void);
    virtual cString SVDRPCommand(const char *Command, const char *Option, int &ReplyCode);
};

#endif
